#define WIN32_LEAN_AND_MEAN 

#include <windows.h> 
#include <stdint.h>
#include <stdlib.h>
#include "small_dll.h"   
#include "vsm.hpp"   

#define default_inp_dly    200
#define cmd_valid   0b10000
#define cmd_read    0b10000
#define cmd_write   0b10100
#define cmd_or      0b11000
#define cmd_and     0b11100
#define cmd_mask    0b11100
#define port_mask   0b00011
#define P2  4
#define _BV(bit) (1 << (bit))

class m8243: public IDSIMMODEL {
public:
    int isdigital(CHAR *pinname);
    void setup(IINSTANCE *inst, IDSIMCKT *dsim);
    void runctrl(RUNMODES mode);
    void actuate(REALTIME time, ACTIVESTATE newstate);
    BOOL indicate(REALTIME time, ACTIVEDATA *data);
    void simulate(ABSTIME time, DSIMMODES mode);
    void callback(ABSTIME time, EVENTID eventid);
private:
    void pin_prog_onchange(ABSTIME time, DSIMMODES mode);
    void pin_cs_onchange(ABSTIME time, DSIMMODES mode);
    uint8_t _fastcall get_port_data(unsigned port);
    void set_port_data(ABSTIME time, uint8_t data, unsigned port);
    void set_port_tristate(ABSTIME time, unsigned port);
    IDSIMPIN *pin_prog;
    IDSIMPIN *pin_cs;
    IDSIMPIN *pin_ports[20];
    IDSIMCKT *ckt;
    long inp_dly;
    unsigned pcmd = 0;
    uint8_t pd[4] = { 0 };
    uint8_t dir = 0;
};

extern "C" IDSIMMODEL __declspec(dllexport) *createdsimmodel(CHAR *device, ILICENCESERVER *ils) {
    ils->authorize(0x80808043);
    return new m8243; 
}

extern "C" void __declspec(dllexport)deletedsimmodel(IDSIMMODEL *model) {
    delete (m8243*)model;
}

void m8243::setup(IINSTANCE *inst, IDSIMCKT *dsim) {
    ckt = dsim;
    unsigned d = (unsigned)inst->getinitval("INP_DLY", 0);
    if (d > 9999) {
        inst->error("m8243: INP_DLY value is too high.");
        return;
    }
    if (d == 0) d = default_inp_dly; else inst->log("m8243: INP_DLY set to %uns.", d);
    inp_dly = d * 1000;
    pin_prog = inst->getdsimpin("$PROG$", true);
    pin_prog->sethandler(this, (PINHANDLERFN)&m8243::pin_prog_onchange);
    pin_cs = inst->getdsimpin("$CS$", true); 
    pin_cs->sethandler(this, (PINHANDLERFN)&m8243::pin_cs_onchange);
    char pstr[4] = { 'P', '4', '0', 0 };
    unsigned i = 0;
    for (;;) {
        pin_ports[i] = inst->getdsimpin(pstr, true);
        pin_ports[i++]->drivetristate(0);
        char c = ++pstr[2];
        if (c == '4') {
            c = ++pstr[1];
            if (c == '3') break;
            if (c == '8') pstr[1] = '2';
            pstr[2] = '0';
        }
    }
}

BOOL m8243::indicate(REALTIME time, ACTIVEDATA *data) {
    return false;
}

int m8243::isdigital(CHAR *pinname) {
    return 1;
}

void m8243::simulate(ABSTIME time, DSIMMODES mode) {
}

#pragma optimize("y", on)
uint8_t m8243::get_port_data(unsigned port) {
    unsigned i = port * 4;
    uint8_t tmp = 0;
    if (!islow(pin_ports[i + 0]->istate())) tmp |= 1;
    if (!islow(pin_ports[i + 1]->istate())) tmp |= 2;
    if (!islow(pin_ports[i + 2]->istate())) tmp |= 4;
    if (!islow(pin_ports[i + 3]->istate())) tmp |= 8;
    return tmp;
}

void m8243::set_port_data(ABSTIME time, uint8_t data, unsigned port) {
    unsigned i = port * 4;
    pin_ports[i + 0]->drivebool(time, data & 1);
    pin_ports[i + 1]->drivebool(time, data & 2);
    pin_ports[i + 2]->drivebool(time, data & 4);
    pin_ports[i + 3]->drivebool(time, data & 8);
}

void m8243::set_port_tristate(ABSTIME time, unsigned port) {
    unsigned i = port * 4;
    pin_ports[i + 0]->drivetristate(time);
    pin_ports[i + 1]->drivetristate(time);
    pin_ports[i + 2]->drivetristate(time);
    pin_ports[i + 3]->drivetristate(time);
}

void m8243::pin_cs_onchange(ABSTIME time, DSIMMODES mode) {
    if (pin_cs->isposedge()) {
        if ((pcmd & cmd_mask) == cmd_read) {
            set_port_tristate(time, P2);
        }
        pcmd = 0;
    }
}
    
void m8243::callback(ABSTIME time, EVENTID eventid) {
    if ((eventid & ~port_mask) != 0x824300) return;
    if (pin_cs->isinactive()) {
        unsigned port = eventid & port_mask;
        uint8_t tmp = get_port_data(port);
        pd[port] = tmp;
        set_port_data(time, tmp, P2);
    }
}

void m8243::pin_prog_onchange(ABSTIME time, DSIMMODES mode) {
    unsigned cmd;
    unsigned port;
    uint8_t tmp;
    if (!pin_cs->isinactive()) return;
    if (pin_prog->isnegedge()) {
        cmd = get_port_data(P2) | cmd_valid;
        pcmd = cmd;
        port = cmd & port_mask;
        cmd &= cmd_mask;
        if (cmd == cmd_read) {           
            if (dir & _BV(port)) {
                dir &= ~_BV(port);
                set_port_tristate(time, port);
            }
            ckt->setcallback(time + inp_dly, this, port + 0x824300);
        }
        return;
    }
    if (pin_prog->isposedge()) {
        cmd = pcmd;
        port = cmd & port_mask;
        cmd &= cmd_mask;
        if (cmd == cmd_read) {
            set_port_tristate(time, P2);
            pcmd = 0;
            return;
        }
        if (cmd) {
            tmp = get_port_data(P2);
            dir |= _BV(port);
            if (cmd == cmd_or) tmp |= pd[port];
            if (cmd == cmd_and) tmp &= pd[port];
            pd[port] = tmp;
            set_port_data(time, tmp, port);
        }
    }
}

void m8243::runctrl(RUNMODES mode) {
}

void m8243::actuate(REALTIME time, ACTIVESTATE newstate) {
}

